#include <stdlib.h>
#include <string.h>

#include <iostream.h>

#include "head.h"

FitsHead::FitsHead(char* raw, int bytes, Memory mem)
{
  cards_ = raw;
  index_ = NULL;
  memory_ = mem;
  ncard_ = 0;
  acard_ = 0;

  valid_ = 0;

  hdu_ = NULL;

  char* c;
  int i;
  for (c = cards_, i=0; c < cards_+bytes; c+=FT_CARDLEN, i++) {
    if (!strncmp(c, "END     ", 8)) {
      ncard_ = i + 1;
      acard_ = ((ncard_+FT_CARDS-1)/FT_CARDS) * FT_CARDS;
      valid_ = 1;

      updateHDU();
      return;
    }
  }
}

// Create Image FitsHead

FitsHead::FitsHead(int width, int height, int bitpix)
{
  cards_ = new char[FT_BLOCK];
  memset(cards_, ' ', FT_BLOCK);
  memcpy(cards_, "END", 3);

  index_ = NULL;
  memory_ = ALLOC;
  ncard_ = 1;
  acard_ = FT_CARDS;

  appendLogical("SIMPLE", 1, "Fits Standard");
  appendInteger("BITPIX", bitpix, "Bits per pixel");
  appendInteger("NAXIS", 2, "Number of axes");
  appendInteger("NAXIS1", width, "Axis Length");
  appendInteger("NAXIS2", height, "Axis Length");

  valid_ = 1;
  hdu_ = NULL;
  updateHDU();
}

FitsHead::FitsHead(const FitsHead& a)
{
  cards_ = new char[a.acard_*FT_CARDLEN];
  memmove(cards_, a.cards_, a.acard_*FT_CARDLEN);
  index_ = NULL;
  memory_ = ALLOC;
  acard_ = a.acard_;
  ncard_ = a.ncard_;

  valid_ = 1;

  hdu_ = NULL;

  updateHDU();
}

FitsHead::~FitsHead()
{
  if (index_)
    delete index_;

  if (hdu_)
    delete hdu_;

  switch (memory_) {
  case ALLOC:
    if (cards_)
      delete [] cards_;
    break;
  case EXTERNAL:
    break;
  }
}

int FitsHead::isImage()
{
  // just look for SIMPLE, if present it may be of value 'F'
  char* xtension = getString("XTENSION");
  char* simple = find("SIMPLE");
  int r = (simple || (xtension && !strcmp(xtension, "IMAGE")));

  delete xtension;
  return r;
}

int FitsHead::isTable()
{
  char* xtension = getString("XTENSION");
  int r = (xtension && (!strcmp(xtension, "TABLE") ||
			!strcmp(xtension, "BINTABLE")));

  delete xtension;
  return r;
}

void FitsHead::updateHDU()
{
  if (hdu_)
    delete hdu_;
  hdu_ = NULL;

  // just find simple, it might be present but of value 'F'
  char* xtension = getString("XTENSION");
  char* simple = find("SIMPLE");

  if (simple || (xtension && !strcmp(xtension, "IMAGE")))
    hdu_ = new FitsImageHDU(this);

  if (xtension && !strcmp(xtension, "TABLE"))
    hdu_ = new FitsAsciiTableHDU(this);

  if (xtension && !strcmp(xtension, "BINTABLE"))
    hdu_ = new FitsBinTableHDU(this);

  delete xtension;
}

int FitsHead::getLogical(const char* name, int def)
{
  char* card = find(name);
  if (card) {
    FitsCard c(card);
    return c.getLogical();
  }
  else
    return def;
}

int FitsHead::getInteger(const char* name, int def)
{
  char* card = find(name);
  if (card) {
    FitsCard c(card);
    return c.getInteger();
  }
  else
    return def;
}

double FitsHead::getReal(const char* name, double def)
{
  char* card = find(name);
  if (card) {
    FitsCard c(card);
    return c.getReal();
  }
  else
    return def;
}

void FitsHead::getComplex(const char* name, double* real, double* img,
			  double rdef, double idef)
{
  char* card = find(name);
  if (card) {
    FitsCard c(card);
    c.getComplex(real, img);
  }
  else {
    *real = rdef;
    *img = idef;
  }
}

char* FitsHead::getString(const char* name)
{
  char* card = find(name);
  if (card) {
    FitsCard c(card);
    return c.getString();
  }
  else
    return NULL;
}

char* FitsHead::getComment(const char* name)
{
  char* card = find(name);
  if (card) {
    FitsCard c(card);
    return c.getComment();
  }
  else
    return NULL;
}

void FitsHead::buildIndex()
{
  if (index_)
    delete index_;

  index_ = new char*[ncard_];
  for (int i=0; i<ncard_; i++)
    index_[i] = cards_ + (i*FT_CARDLEN);

  qsort(index_, ncard_, sizeof(char**), ft_compare);
}

char* FitsHead::find(const char* key)
{
  return (index_ ? findIndex(key) : findSeq(key));
}

char* FitsHead::findIndex(const char* key)
{
  if (key == NULL)
    return NULL;

  if (!index_)
    buildIndex();

  char** base = index_;
  int length = ncard_;

  int lo = -1;
  int hi = length;
  int cut = length/2;

  while (hi-lo > 1) {
    int i; 
    if (!(i = ft_compare(&key, &base[cut])) ) {
      return base[cut];
    }
 
    if (i < 0) {
      hi = cut;
      cut = (lo+hi)/2;
    }
    else { 
      lo = cut;
      cut = (lo+hi)/2;
    } 
  }

  if (!strncmp(key, base[cut], 5))
    return base[cut];

  return NULL;
}
 
char* FitsHead::findSeq(const char* key)
{
  if (key == NULL)
    return NULL;

  char k[8];
  memset(k,' ',8);
  int len = strlen(key);
  memcpy(k, key, (len>8 ? 8 : len));

  for (char* card = cards_; card != &cards_[ncard_*FT_CARDLEN]; card+=FT_CARDLEN)
    if (!strncmp(k, card, 8))
      return card;

  return NULL;
}

char* FitsHead::setLogical(const char* name, int value, const char* comm)
{
  char* card = find(name);
  if (card)
    FitsCard(card).setLogical(value, comm);

  return card;
}

char* FitsHead::setInteger(const char* name, int value, const char* comm)
{
  char* card = find(name);
  if (card)
    FitsCard(card).setInteger(value, comm);

  return card;
}

char* FitsHead::setReal(const char* name, double value, int prec, const char* comm)
{
  char* card = find(name);
  if (card)
    FitsCard(card).setReal(value, prec, comm);

  return card;
}

char* FitsHead::setComplex(const char* name, double real, double img, int prec,
			   const char* comm)
{
  char* card = find(name);
  if (card)
    FitsCard(card).setComplex(real, img, prec, comm);

  return card;
}

char* FitsHead::setString(const char* name, const char* value, const char* comm)
{
  char* card = find(name);
  if (card)
    FitsCard(card).setString(value, comm);

  return card;
}

char* FitsHead::setComment(const char* name, const char* value)
{
  char* card = find(name);
  if (card)
    FitsCard(card).setComment(value);

  return card;
}

char* FitsHead::cardins(char* card, char* here)
{
  char* where = here ? here : cards_+((ncard_-1)*FT_CARDLEN);

  if (ncard_+1 > acard_) {
    // not implemented yet
  }
    
  memmove(where+FT_CARDLEN, where, (cards_+(ncard_*FT_CARDLEN))-where);
  memmove(where, card, FT_CARDLEN);
  ncard_++;

  if (index_)
    buildIndex();

  return where;
}

char* FitsHead::carddel(char* card)
{
  return card;
}

char* FitsHead::insertLogical(const char* name, int value, const char* comm, 
			      const char* here)
{
  FitsCard key;
  key.setLogical(name, value, comm);
  return cardins(key.card(), (char*)here);
}

char* FitsHead::insertInteger(const char* name, int value, const char* comm, 
			      const char* here)
{
  FitsCard key;
  key.setInteger(name, value, comm);
  return cardins(key.card(), (char*)here);
}

char* FitsHead::insertReal(const char* name, double value, int prec, const char* comm,
			   const char* here)
{
  FitsCard key;
  key.setReal(name, value, prec, comm);
  return cardins(key.card(), (char*)here);
}

char* FitsHead::insertComplex(const char* name, double real, double img, 
			      int prec, const char* comm, const char* here)
{
  FitsCard key;
  key.setComplex(name, real, img, prec, comm);
  return cardins(key.card(), (char*)here);
}

char* FitsHead::insertString(const char* name, const char* value, const char* comm,
			     const char* here)
{
  FitsCard key;
  key.setString(name, value, comm);
  return cardins(key.card(), (char*)here);
}

char* FitsHead::insertComment(const char* name, const char* value, const char* here)
{
  FitsCard key;
  key.setComment(name, value);
  return cardins(key.card(), (char*)here);
}

int ft_compare(const void* a, const void* b)
{
  if (!strncmp((char*)a, (char*)b, 5)) {
    int ax = atoi((const char*)((char*)a)[5]);
    int bx = atoi((const char*)((char*)b)[5]);

    if (ax && bx) {
      if (ax <  bx)
	return -1;
      if (ax == bx)
	return  0;
      if (ax >  bx)
	return  1;
    }
  }

  return strncmp((char*)a, (char*)b, 8);
}
