#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <limits.h>

#include <iostream.h>
#include <strstream.h>

#include "fitsimage.h"
#include "frame.h"
#include "util.h"

#include "mmap.h"
#include "alloc.h"
#include "channel.h"
#include "share.h"
#include "iis.h"
#include "hist.h"

// we use this matrix to tie mosaics together
// the first loaded is used as the 'anchor'

Matrix FitsImage::baseDetectorToData;

FitsImage::FitsImage(Frame* p, int m, int fm)
{
  parent = p;
  valid = 0;
  objectName = NULL;
  fileName = NULL;
  fullFileName = NULL;
  fits = NULL;
  image = NULL;
  data = NULL;

  isHist_ = 0;

  width_ = 0;
  height_ = 0;

  rotation = 0;
  zoom = Vector(1,1);
  orientation = Frame::NORMAL;

  keyLTMV = 0;
  keyATMV = 0;
  keyDTMV = 0;

  keySEC = 0;
  keySIZE = 0;

  mosaicMode = m;
  firstMosaic = fm;

  wcs = NULL;
  next_ = NULL;

  iisMode = 0;
  iiszt = 0;
}

FitsImage::~FitsImage()
{
  if (objectName)
    delete objectName;

  if (fileName)
    delete fileName;

  if (fullFileName)
    delete fullFileName;

  if (data)
    delete data;

  if (isHist_ && image)
    delete image;

  if (wcs)
    free(wcs);
}

FitsImageMaster::~FitsImageMaster()
{
  if (fits)
    delete fits;
}

FitsImageSlave::FitsImageSlave(Frame* p, const char* fn, int m, int fm, 
			       FitsFile* f) : FitsImage(p, m, fm)
{
  setFileName(fn);
  fits = f;
  process();
}

FitsImageFitsAlloc::FitsImageFitsAlloc(Frame* p, const char* fn, int m, int fm)
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  fits = new FitsFitsAlloc(fn, FitsFile::RELAX);
  process();
}

FitsImageFitsChannel::FitsImageFitsChannel(Frame* p, Tcl_Interp* interp, 
					   const char* ch, const char* fn, 
					   int m, int fm)
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  fits = new FitsFitsChannel(interp, ch, fn, FitsFile::RELAX);
  process();
}

FitsImageFitsMMap::FitsImageFitsMMap(Frame* p, const char* fn, int m, int fm) 
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  fits = new FitsFitsMMap(fn, FitsFile::RELAX);
  process();
}

FitsImageFitsShare::FitsImageFitsShare(Frame* p, Frame::ShmType type, 
				       int id, int size, const char* fn,
				       int m, int fm)
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  switch (type) {
  case Frame::SHMID:
    fits = new FitsFitsShareID(id, size, fn, FitsFile::RELAX);
    break;
  case Frame::KEY:
    fits = new FitsFitsShareKey(id, size, fn, FitsFile::RELAX);
    break;
  }
  process();
}

FitsImageArrAlloc::FitsImageArrAlloc(Frame* p, const char* fn, int m, int fm)
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  fits = new FitsArrAlloc(fn);
  process();
}

FitsImageArrChannel::FitsImageArrChannel(Frame* p, Tcl_Interp* interp, 
					 const char* ch, const char* fn, 
					 int m, int fm) 
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  fits = new FitsArrChannel(interp, ch, fn);
  process();
}

FitsImageArrMMap::FitsImageArrMMap(Frame* p, const char* fn, int m, int fm) 
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  fits = new FitsArrMMap(fn);
  process();
}

FitsImageArrShare::FitsImageArrShare(Frame* p, Frame::ShmType type, int id,
				     int size, const char* fn, int m, int fm) 
  : FitsImageMaster(p, m, fm)
{
  setFileName(fn);
  switch (type) {
  case Frame::SHMID:
    fits = new FitsArrShareID(id, size, fn);
    break;
  case Frame::KEY:
    fits = new FitsArrShareKey(id, size, fn);
    break;
  }
  process();
}

FitsImageIIS::FitsImageIIS(Frame* p, int w, int h) : FitsImageMaster(p, 0, 0)
{
  fits = new FitsIIS(w, h);
  process();
  iisMode = 1;
}

FitsImageMosaicMMap::FitsImageMosaicMMap(Frame* p, const char* fn)
  : FitsImageMaster(p, 1, 1)
{
  fits = new FitsMosaicMMap(fn);
  processMosaic(fn);
}

FitsImageMosaicShare::FitsImageMosaicShare(Frame* p, Frame::ShmType type,
					   int id, int size, const char* fn) 
  : FitsImageMaster(p, 1, 1)
{
  switch (type) {
  case Frame::SHMID:
    fits = new FitsMosaicShareID(id, size);
    break;
  case Frame::KEY:
    fits = new FitsMosaicShareKey(id, size);
    break;
  }
  processMosaic(fn);
}

FitsImageMosaicNextMap::FitsImageMosaicNextMap(Frame* p, const char* fn, 
					       FitsFile* prev)
  : FitsImageMaster(p, 1, 0)
{
  fits = new FitsMosaicNextMap(prev);
  processMosaic(fn);
}

FitsImageMosaicAlloc::FitsImageMosaicAlloc(Frame* p, const char* fn)
  : FitsImageMaster(p, 1, 1)
{
  fits = new FitsMosaicAlloc(fn);
  processMosaic(fn);
}

FitsImageMosaicChannel::FitsImageMosaicChannel(Frame* p, Tcl_Interp* interp,
					       const char* ch, const char* fn)
					       : FitsImageMaster(p, 1, 1)
{
  fits = new FitsMosaicChannel(interp, ch);
  processMosaic(fn);
}

FitsImageMosaicNextAlloc::FitsImageMosaicNextAlloc(Frame* p, const char* fn,
						   FitsFile* prev)
  : FitsImageMaster(p,1,0)
{
  fits = new FitsMosaicNextAlloc(prev);
  processMosaic(fn);
}

FitsImageMosaicNextChannel::FitsImageMosaicNextChannel(Frame* p, 
						       const char* fn, 
						       FitsFile* prev)
  : FitsImageMaster(p,1,0)
{
  fits = new FitsMosaicNextChannel(prev);
  processMosaic(fn);
}

void FitsImageIIS::iisErase()
{
  if (fits)
    ((FitsIIS*)fits)->erase();
}

const char* FitsImageIIS::iisGet(int x, int y)
{
  if (fits)
    return ((FitsIIS*)fits)->get(x,y);
}

void FitsImageIIS::iisSet(int x, int y, const char* d, int l)
{
  if (fits)
    ((FitsIIS*)fits)->set(x,y,d,l);
}

void FitsImageIIS::iisWCS(const Matrix& m, const Vector& z, int zt)
{
  dataToImage = Translate(-.5,-.5) * m;
  imageToData = dataToImage.invert();
  iisz = z;
  iiszt = zt;
}

void FitsImage::process()
{
  if (!fits->isValid())
    goto error;

  if (fits->isImage()) {
    image = fits;
    isHist_ = 0;

    FitsHead* head = image->head();
    width_ = head->naxis(1);
    height_ = head->naxis(2);
    center = Vector(width_, height_)/2;

    load();
    valid = 1;
    return;
  }
  else if (fits->isTable()) {

    isHist_ = 1;
    initHist();
    return;
  }

error:
  delete fits;
  fits = NULL;
  image = NULL;
  isHist_ = 0;
  valid = 0;
}

void FitsImage::processMosaic(const char* fn)
{
  if (fits->find("EXTNAME")) {
    char lbuf[256];
    {
      ostrstream str(lbuf,256);
      str << fn << '[' << fits->getString("EXTNAME") << ']' << ends;
      fullFileName = dupstr(lbuf);
    }
    {
      char* m = root(fn);
      ostrstream str(lbuf,256);
      str << m << '[' << fits->getString("EXTNAME") << ']' << ends;
      delete m;
      fileName = dupstr(lbuf);
    }
  }
  else {
    fullFileName = dupstr(fn);
    fileName = root(fn);
  }

  if (fits->isValid()) {
    if (fits->isImage()) {
      image = fits;
      isHist_ = 0;

      FitsHead* head = image->head();
      width_ = head->naxis(1);
      height_ = head->naxis(2);
      center = Vector(width_, height_)/2;

      load();
      valid = 1;
      return;
    }
    else if (fits->isTable()) {
      // just return, don't try to bin it.
      valid = 1;
      return;
    }
  }

  // if we get this far, we have an error

error:
  delete fits;
  fits = NULL;
  image = NULL;
  isHist_ = 0;
  valid = 0;
}

void FitsImage::setFileName(const char* fn)
{
  fullFileName = dupstr(fn);
  fileName = root(fn);
}

char* FitsImage::root(const char* fn)
{
  if (fn) {
    const char* ptr = fn;           // init the ptr
    while(*ptr++);                  // walk it forward to end of string
    ptr--;                          // backup one
    while(*ptr != '/' && ptr != fn) // walk it backward til last / or beginning
      ptr--;
    if (*ptr == '/')                // step it over the last '/'
      ptr++;
    return dupstr(ptr);             // got it!
  }
  else
    return NULL;
}

int FitsImage::saveFitsFile(const char* fn)
{
  if (image)
    return image->saveFitsFile(fn);
  else
    return 0;
}

int FitsImage::saveFitsChannel(Tcl_Interp* interp, const char* ch)
{
  if (image)
    return image->saveFitsChannel(interp, ch);
  else
    return 0;
}

void FitsImage::updateMatrices(Matrix& refToUser, Matrix& userToWidget, 
			       Matrix& widgetToCanvas, Matrix& userToPanner)
{
  // basic matrices

  // if in mosaic mode, we need to flip the image, based on orientation
  // which is set via the ATM/V keywords when the file is parsed.

  if (mosaicMode && keySEC && keySIZE) {
    Matrix flip;
    switch (orientation) {
    case Frame::X:
      flip = FlipX();
      break;
    case Frame::Y:
      flip = FlipY();
      break;
    case Frame::XY:
      flip = FlipXY();
      break;
    }

    dataToRef =
      Translate(-center) *
      flip *
      Rotate(rotation) *
      Scale(zoom) *
      Translate(center) *
      Translate(origin);

    refToData = dataToRef.invert();
  }
  else {
    dataToRef.identity();
    refToData.identity();
  }

  // derived matrices

  physicalToData = physicalToImage * imageToData;
  dataToPhysical = physicalToData.invert();

  physicalToRef = physicalToImage * imageToData * dataToRef;
  refToPhysical = physicalToRef.invert();

  physicalToUser = physicalToImage * imageToData * dataToRef * refToUser;
  userToPhysical = physicalToUser.invert();

  physicalToCanvas = physicalToImage * imageToData * dataToRef * refToUser * 
    userToWidget * widgetToCanvas;
  canvasToPhysical = physicalToCanvas.invert();

  physicalToPanner = physicalToImage * imageToData * dataToRef * refToUser * 
    userToPanner;
  pannerToPhysical = physicalToPanner.invert();

  imageToRef = imageToData * dataToRef;
  refToImage = imageToRef.invert();

  imageToUser = imageToData * dataToRef * refToUser;
  userToImage = imageToUser.invert();

  imageToWidget = imageToData * dataToRef *  refToUser * userToWidget;
  widgetToImage = imageToWidget.invert();

  imageToCanvas = imageToData * dataToRef *  refToUser *
    userToWidget *  widgetToCanvas;
  canvasToImage = imageToCanvas.invert();

  imageToPanner = imageToData * dataToRef * refToUser * userToPanner;
  pannerToImage = imageToPanner.invert();

  dataToUser = dataToRef * refToUser;
  userToData = dataToUser.invert();

  dataToWidget = dataToRef *  refToUser *  userToWidget;
  widgetToData = dataToWidget.invert();

  dataToCanvas = dataToRef *  refToUser *  userToWidget * widgetToCanvas;
  canvasToData = dataToCanvas.invert();

  dataToPanner = dataToRef * refToUser * userToPanner;
  pannerToData = dataToPanner.invert();

  amplifierToRef = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef;
  refToAmplifier = amplifierToRef.invert();

  amplifierToUser = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser;
  userToAmplifier = amplifierToUser.invert();

  amplifierToCanvas = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser *
    userToWidget * 
    widgetToCanvas;
  canvasToAmplifier = amplifierToCanvas.invert();

  amplifierToPanner = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser * 
    userToPanner;
  pannerToAmplifier = amplifierToPanner.invert();

  detectorToRef = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef;
  refToDetector = detectorToRef.invert();

  detectorToUser = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser;
  userToDetector = detectorToUser.invert();

  detectorToCanvas = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser *
    userToWidget * 
    widgetToCanvas;
  canvasToDetector = detectorToCanvas.invert();

  detectorToPanner = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser * 
    userToPanner;
  pannerToDetector = detectorToPanner.invert();
}

const char* FitsImage::getValue(const Vector& v)
{
  if (!isIIS() || (iiszt != 1))
    return data->getValue(v);
  else {
    double value = data->getValueDouble(v);

    ostrstream str(buf,32);
    if (value <= 1)
      str << '<' << iisz[0] << ends;
    else if (value >= 200)
      str << '>' << iisz[1] << ends;
    else
      str << ((value-1) * (iisz[1]-iisz[0]))/199 + iisz[0] << ends;

    return buf;
  }
}

void FitsImage::initHist()
{
  // just in case
  if (!fits->isTable())
    return;

  // make sure we have cols to bin on
  if (!fits->pBinX())
    fits->setpBinX("X");

  if (!fits->pBinY())
    fits->setpBinY("Y");

  // determine buffersize
  //  width_ = height_ = parent->blockBufferSize;
  setHistSize();
  center = Vector(width_, height_)/2;

  // force to a block boundary, then translate to center of block cell
  histCursor = (getHistCenter()/parent->blockFactor).ceil() *
    parent->blockFactor;
  histCursor += Vector(0,0) *
    Translate(-.5,-.5) * Scale(parent->blockFactor) * Translate(.5,.5);

  // build the matrix

  binToData = 
    Translate(-histCursor) *
    Scale(1./parent->blockFactor) *
    Translate(center);
  dataToBin = binToData.invert();

  image = new FitsHist(fits, width_, height_, binToData,
		       parent->blockFunction, parent->blockFactor);

  if (image->isValid()) {
    load();
    valid = 1;
  }
  else {
    delete fits;
    fits = NULL;
    image = NULL;
    isHist_ = 0;
    valid = 0;
  }
}

Matrix FitsImage::updateHist()
{
  if (isHist_)
    return nextHist(getHistCenter(),GRAVITY);
  else
    return Matrix();
}

Matrix FitsImage::updateHist(Vector v)
{
  if (isHist_)
    return nextHist(v,NOGRAVITY);
  else
    return Matrix();
}

Matrix FitsImage::updateHistUser(Vector v)
{
  if (isHist_)
    return nextHist(v * userToData * dataToBin, GRAVITY);
  else
    return Matrix();
}

Matrix FitsImage::nextHist(Vector c, Gravity g)
{
  // cursor is in bin (physical) coords
  //  width_ = height_ = parent->blockBufferSize;
  setHistSize();
  center = Vector(width_, height_)/2;

  if (g == NOGRAVITY)
    histCursor = c;
  else {
    // force to a block boundary, then translate to center of block cell
    histCursor = ((c/parent->blockFactor).ceil())*parent->blockFactor;
    histCursor += Vector(0,0) *
      Translate(-.5,-.5) * Scale(parent->blockFactor) * Translate(.5,.5);
  }

  // save for current markers

  Matrix oldDataToBin = dataToBin;

  // new xform from original (virtual) image to binned image

  binToData =
    Translate(-histCursor) *
    Scale(1./parent->blockFactor) *
    Translate(center);
  dataToBin = binToData.invert();

  parent->cursor = center;

  if (image)
    delete image;
  image = new FitsHist(fits, width_, height_, binToData,
		       parent->blockFunction, parent->blockFactor);

  if (image->isValid()) {
    load();
    valid = 1;
    // I'm not sure why we need a half pixel translate at this point
    // but it works
    Matrix mm = 
      Translate(-.5,-.5) *
      refToData * 
      oldDataToBin * 
      binToData * 
      dataToRef *
      Translate(.5,.5);
    return mm;
  }
  else {
    delete fits;
    fits = NULL;
    image = NULL;
    isHist_ = 0;
    valid = 0;
  }
}

void FitsImage::setHistSize()
{
  Vector s = getHistDim()/parent->blockFactor;

  width_ = (int)(s[0]<parent->blockBufferSize? s[0] : parent->blockBufferSize);
  height_= (int)(s[1]<parent->blockBufferSize? s[1] : parent->blockBufferSize);
}

Vector FitsImage::getHistDim()
{
  // assumes we aready have our columns

  FitsTableHDU* hdu = (FitsTableHDU*)(fits->head())->hdu();
  Vector xd = hdu->dimension(fits->pBinX());
  Vector yd = hdu->dimension(fits->pBinY());
  return Vector(xd[1]-xd[0]+1,yd[1]-yd[0]+1);
}

Vector FitsImage::getHistCenter()
{
  // assumes we aready have our columns

  FitsTableHDU* hdu = (FitsTableHDU*)(fits->head())->hdu();
  Vector xd = hdu->dimension(fits->pBinX());
  Vector yd = hdu->dimension(fits->pBinY());
  return Vector((xd[1]-xd[0]+1)/2+xd[0], (yd[1]-yd[0]+1)/2+yd[0]);
}

void FitsImage::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  dataToMagnifier = dataToRef * refToMagnifier;
  magnifierToData = dataToMagnifier.invert();
}

Vector FitsImage::getCenter()
{
  if (mosaicMode && keySEC && keySIZE) {
    Vector r = detsize.center() * baseDetectorToData;
    if (DebugMosaic)
      cerr << "detsize.center: " << r.round() << endl;
    return r.round();
  }
  else
    return center;
}

Vector FitsImage::getSize()
{
  if (mosaicMode && keySEC && keySIZE) {
    BBox bb = detsize * baseDetectorToData;
    return bb.size();
  }
  else
    return Vector(width_, height_);
}

void FitsImage::load()
{
  FitsHead* head = image->head();

  // process keywords

  processKeywords();

  // create a valid FitsData

  switch (head->bitpix()) {
  case 8: 
    data = new FitsDatam<unsigned char>(image);
    break;
  case 16:
    data = new FitsDatam<short>(image);
    break;
  case -16:
    data = new FitsDatam<unsigned short>(image);
    break;
  case 32:
    data = new FitsDatam<BITPIX32>(image);
    break;
  case -32:
    data = new FitsDatam<float>(image);
    break;
  case -64:
    data = new FitsDatam<double>(image);
    break;
  }

  // init WCS

  hlength(head->cards(),head->headbytes());
  wcs = wcsinit(head->cards());

  if (DebugWCS)

    if (hasWCS()) {
      cerr << "wcs->rot=" << wcs->rot << endl;
      cerr << "wcs->equinox=" << wcs->equinox << endl;
      cerr << "wcs->epoch=" << wcs->epoch << endl;
      cerr << "wcs->longpole=" << wcs->longpole << endl;
      cerr << "wcs->latpole=" << wcs->latpole << endl;

      {
	for (int i=0; i<2; i++)
	  cerr << "wcs->crpix[" << i << "]=" << wcs->crpix[i] << endl;
      }
      {
	for (int i=0; i<2; i++)
	  cerr << "wcs->crval[" << i << "]=" << wcs->crval[i] << endl;
      }
      {
	for (int i=0; i<2; i++)
	  cerr << "wcs->cdelt[" << i << "]=" << wcs->cdelt[i] << endl;
      }
      {
	for (int i=0; i<4; i++)
	  cerr << "wcs->cd[" << i << "]=" << wcs->cd[i] << endl;
      }
      {
	for (int i=0; i<4; i++)
	  cerr << "wcs->pc[" << i << "]=" << wcs->pc[i] << endl;
      }
      cerr << "wcs->imrot=" << wcs->imrot << endl;
      cerr << "wcs->pa_north=" << wcs->pa_north << endl;
      cerr << "wcs->pa_east=" << wcs->pa_east << endl;
      cerr << "wcs->imflip=" << wcs->imflip << endl;
      cerr << "wcs->prjcode=" << wcs->prjcode << endl;

      cerr << "wcs->coorflip=" << wcs->coorflip << endl;
      cerr << "wcs->wcsproj=" << wcs->wcsproj << endl;

      {
	for (int i=0; i<2; i++)
	  cerr << "wcs->ctype[" << i << "]=" << wcs->ctype[i] << endl;
      }
      cerr << "wcs->c1type=" << wcs->c1type << endl;
      cerr << "wcs->c2type=" << wcs->c2type << endl;
      cerr << "wcs->ptype=" << wcs->ptype << endl;
      cerr << "wcs->radecsys=" << wcs->radecsys << endl;
      cerr << "wcs->radecout=" << wcs->radecout << endl;
      cerr << "wcs->syswcs=" << wcs->syswcs << endl;
    }
}

char* FitsImage::getHeader()
{
  if (image) {
    FitsHead* head = image->head();
    if (head)
      return head->cards();
  }
  return NULL;
}

int FitsImage::getNCards()
{
  if (image) {
    FitsHead* head = image->head();
    if (head)
      return head->ncard();
  }
  return 0;
}

char* FitsImage::displayHeader()
{
  // always return the header to the native fits ext, not the image
  FitsHead* head = fits->head();
  //  FitsHead* head = image->head();
  int size = head->ncard() * FT_CARDLEN;
  char* lbuf = new char[size+1];
  memcpy(lbuf, head->cards(), size);

  for(int i=0; i<head->ncard(); i++){
    lbuf[i*FT_CARDLEN+79] = '\n';
  }

  lbuf[size] = '\0';
  return lbuf;
}

void FitsImage::parseSection(char* lbuf, Vector* v1, Vector* v2)
{
  double x1, y1, x2, y2;
  char d; // dummy char
  istrstream str(lbuf);

  str >> d >> (*v1)[0] >> d >> (*v2)[0] >> d >> (*v1)[1] >> d >> (*v2)[1] >> d;
}

void FitsImage::processKeywords()
{
  // Process OBJECT keyword

  objectName = image->getString("OBJECT");

  // Image to Data

  imageToData = Translate(-.5, -.5);
  dataToImage = imageToData.invert();

  // Physical to Image (LTM/LTV keywords)

  if (image->find("LTM1_1") ||
      image->find("LTM1_2") ||
      image->find("LTM2_1") ||
      image->find("LTM2_2") ||
      image->find("LTV1") ||
      image->find("LTV2"))
    keyLTMV = 1;

  double ltm11 = image->getReal("LTM1_1", 1);
  double ltm12 = image->getReal("LTM1_2", 0);
  double ltm21 = image->getReal("LTM2_1", 0);
  double ltm22 = image->getReal("LTM2_2", 1);

  double ltv1 = image->getReal("LTV1", 0);
  double ltv2 = image->getReal("LTV2", 0);

  physicalToImage = Matrix(ltm11, ltm12, ltm21, ltm22, ltv1, ltv2);
  imageToPhysical = physicalToImage.invert();

  // CDD to Detector (DTM/DTV keywords)

  if (image->find("DTM1_1") ||
      image->find("DTM1_2") ||
      image->find("DTM2_1") ||
      image->find("DTM2_2") ||
      image->find("DTV1") ||
      image->find("DTV2"))
    keyDTMV = 1;

  double dtm11 = image->getReal("DTM1_1", 1);
  double dtm12 = image->getReal("DTM1_2", 0);
  double dtm21 = image->getReal("DTM2_1", 0);
  double dtm22 = image->getReal("DTM2_2", 1);

  double dtv1 = image->getReal("DTV1", 0);
  double dtv2 = image->getReal("DTV2", 0);

  physicalToDetector = Matrix(dtm11, dtm12, dtm21, dtm22, dtv1, dtv2);
  detectorToPhysical = physicalToDetector.invert();

  // Physical to Amplifier (ATM/ATV keywords)

  if (image->find("ATM1_1") ||
      image->find("ATM1_2") ||
      image->find("ATM2_1") ||
      image->find("ATM2_2") ||
      image->find("ATV1") ||
      image->find("ATV2"))
    keyATMV = 1;

  double atm11 = image->getReal("ATM1_1", 1);
  double atm12 = image->getReal("ATM1_2", 0);
  double atm21 = image->getReal("ATM2_1", 0);
  double atm22 = image->getReal("ATM2_2", 1);

  double atv1 = image->getReal("ATV1", 0);
  double atv2 = image->getReal("ATV2", 0);

  physicalToAmplifier = Matrix(atm11, atm12, atm21, atm22, atv1, atv2);
  amplifierToPhysical = physicalToAmplifier.invert();

  if (DebugMosaic) {
    cerr << endl;
    cerr << fileName << endl;
    cerr << "ATM/V: " << physicalToAmplifier << endl;
    cerr << "ATM/V-1: " << amplifierToPhysical << endl;
    cerr << "DTM/V: " << physicalToDetector << endl;
    cerr << "DTM/V-1: " << detectorToPhysical << endl;
    cerr << "LTM/V: " << physicalToImage << endl;
    cerr << "LTM/V-1: " << imageToPhysical << endl;
  }

  // Determine origin for Mosaics

  if (mosaicMode) {
    // find size

    char* size = image->getString("DETSIZE");
    if (size && *size) {
      Vector v1,v2;
      parseSection(size,&v1,&v2);
      detsize = BBox(v1,v2);
      keySIZE = 1;
      delete size;
    }

    if (DebugMosaic)
      cerr << "detsize: " << detsize << endl;

    // DETSEC

    keySEC = 0;
    char* det =  image->getString("DETSEC");
    if (det && *det) {
      Vector v1,v2;
      parseSection(det,&v1,&v2);
      delete det;

      BBox bb(v1,v2);

      // if this is the first file, we use it to calculate a base matrix
      // used for calculating a scaling/translation in 
      // data coords based on detector coords.

      if (firstMosaic)
	baseDetectorToData = Scale(fabs(detectorToPhysical[0][0]),
				   fabs(detectorToPhysical[1][1])) *
	  Scale(fabs(physicalToImage[0][0]),fabs(physicalToImage[1][1])) * 
	  imageToData;

      origin = bb.ll * baseDetectorToData;

      if (DebugMosaic) {
	cerr << "DETSEC: " << bb << endl;
	cerr << "baseDetectorToData: " << baseDetectorToData << endl;
	cerr << "origin: " << origin << endl;
      }

      keySEC = 1;
    }    

    // AMPSEC
    // all we care about is if AMPSEC is flipped
    // if the amp reads reverse but the data is flip before writing, both
    // AMPSEC and ATM will be neg and we don't want to flip

    char* amp =  image->getString("AMPSEC");
    Vector v1(0,0),v2(1,1);
    if (amp && *amp) {
      parseSection(amp,&v1,&v2);
      delete amp;
    }

    // determine orientation based on ATM/ATV and AMPSEC
    // AMPSEC may not be present

    int xx = !((physicalToAmplifier[0][0]<0) && (v2[0]>v1[0])) ||
      ((physicalToAmplifier[0][0]>0) && (v2[0]<v1[0]));
    int yy = !((physicalToAmplifier[1][1]<0) && (v2[1]>v1[1])) ||
      ((physicalToAmplifier[1][1]>0) && (v2[1]<v1[1]));

    if (xx && yy)
      orientation = Frame::NORMAL;
    else if (!xx && yy)
      orientation = Frame::X;
    else if (!xx && !yy)
      orientation = Frame::XY;
    else if (xx && !yy)
      orientation = Frame::Y;
  }
}

int FitsImage::hasWCS()
{
  return (wcs && iswcs(wcs)) ? 1 : 0;
}

int FitsImage::hasWCSEqu()
{
  switch (wcsToCoordSystem(getradecsys(wcs))) {
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::GALACTIC:
  case Frame::ECLIPTIC:
  case Frame::ICRS:
    return 1;
  default:
    return 0;
  }
}

int FitsImage::hasWCSLinear()
{
  if (wcsToCoordSystem(getradecsys(wcs)) == Frame::LINEAR)
    return 1;
  else
    return 0;
}

char* FitsImage::pix2wcs(Vector in, Frame::CoordSystem sys, 
			 Frame::CoordFormat format, char* lbuf, int len)
{
  if (hasWCS()) {
    wcsoutinit(wcs, (char*)coordSystemToWCS(sys));
    switch (format) {
    case Frame::PIXELS:
    case Frame::DEGREES:
      setwcsdeg(wcs,1);
      wcs->ndec = 5;
      break;
    case Frame::SEXAGESIMAL:
      setwcsdeg(wcs,0);
      wcs->ndec = 3;
      break;
    }
    setwcslin(wcs,2);
    pix2wcst(wcs, in[0], in[1], lbuf, len);
  }

  return lbuf;
}

Vector FitsImage::pix2wcs(Vector in)
{
  if (hasWCS()) {
    wcsoutinit(wcs, getradecsys(wcs));
    double x,y;
    ::pix2wcs(wcs, in[0], in[1], &x, &y);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::pix2wcs(Vector in, Frame::CoordSystem sys)
{
  if (hasWCS()) {
    wcsoutinit(wcs, (char*)coordSystemToWCS(sys));
    double x,y;
    ::pix2wcs(wcs, in[0], in[1], &x, &y);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2pix(Vector in)
{
  if (hasWCS()) {
    wcsininit(wcs, getradecsys(wcs));
    double x,y;
    int off;
    ::wcs2pix(wcs, in[0], in[1], &x, &y, &off);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2pix(Vector in, Frame::CoordSystem sys)
{
  if (hasWCS()) {
    wcsininit(wcs, (char*)coordSystemToWCS(sys));
    double x,y;
    int off;
    ::wcs2pix(wcs, in[0], in[1], &x, &y, &off);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2wcs(Vector in, Frame::CoordSystem sysi, 
			  Frame::CoordSystem syso)
{
  Vector r = in;
  if (hasWCS())
    wcscon(wcscsys((char*)coordSystemToWCS(sysi)), 
	   wcscsys((char*)coordSystemToWCS(syso)), 
	   wcs->equinox, 0., &r[0], &r[1], wcs->epoch);

  return r;
}

Vector FitsImage::wcs2pixDist(Vector in)
{
  if (hasWCS()) {
    Vector cd = getWCScdelt();
    return Vector(in[0]/cd[0], in[1]/cd[1]).abs();
  }
  else
    return in;
}

Vector FitsImage::pix2wcsDist(Vector in)
{
  if (hasWCS()) {
    Vector cd = getWCScdelt();
    return Vector(in[0]*cd[0], in[1]*cd[1]).abs();
  }
  else
    return in;
}

double FitsImage::wcsdist(Vector a, Vector b)
{
  if (hasWCS())
    return ::wcsdist(a[0],a[1],b[0],b[1]);
  else
    return 0;
}

int FitsImage::getWCSFlip()
{
  if (hasWCS())
    return wcs->coorflip;
  else
    return 0;
}

Frame::CoordSystem FitsImage::getWCSCoordSystem()
{
  return wcsToCoordSystem(getradecsys(wcs));
}

const char* FitsImage::getWCSSystem()
{
  if (hasWCS())
    return coordSystemToWCS(getWCSCoordSystem());
  else
    return NULL;
}

Vector FitsImage::getWCScrval()
{
  if (hasWCS()) {
    if (!wcs->coorflip)
      return Vector(wcs->crval[0], wcs->crval[1]);
    else
      return Vector(wcs->crval[1], wcs->crval[0]);
  }
  else
    return Vector();
}

Vector FitsImage::getWCScrpix()
{
  if (hasWCS())
      return Vector(wcs->crpix[0], wcs->crpix[1]);
  else
    return Vector();
}

Vector FitsImage::getWCScdelt()
{
  if (hasWCS()) {
    if (!wcs->coorflip)
      return Vector(wcs->cdelt[0], wcs->cdelt[1]);
    else
      return Vector(wcs->cdelt[1], wcs->cdelt[0]);
  }
  else
    return Vector();
}

const char* FitsImage::coordSystemToWCS(Frame::CoordSystem sys)
{
  switch(sys) {
  case Frame::EQUATORIAL_FK4:
    return wcsName[0];
  case Frame::EQUATORIAL_FK5:
    return wcsName[1];
  case Frame::ICRS:
    return wcsName[2];
  case Frame::GALACTIC:
    return wcsName[3];
  case Frame::ECLIPTIC:
    return wcsName[4];
  case Frame::LINEAR:
    return wcsName[5];
  default:
    return NULL;
  }
}

Frame::CoordSystem FitsImage::wcsToCoordSystem(const char* str)
{
  if (!strncmp("FK4", str, 3))
    return Frame::EQUATORIAL_FK4;
  else if (!strncmp("FK5", str, 3))
    return Frame::EQUATORIAL_FK5;
  else if (!strncmp("ICR", str, 3))
    return Frame::ICRS;
  else if (!strncmp("GAL", str, 3))
    return Frame::GALACTIC;
  else if (!strncmp("ECL", str, 3))
    return Frame::ECLIPTIC;
  else if (!strncmp("LIN", str, 3))
    return Frame::LINEAR;
  else {
    cerr << "Frame Internal Error: Unknown WCS "<< str 
	 << " Assuming LINEAR." << endl;
    return Frame::LINEAR;
  }
}

double FitsImage::getWCSRotation()
{
  if (hasWCS())
    return -degToRad(wcs->imrot);
  else
    return 0;
}

Frame::Orientation FitsImage::getWCSOrientation()
{
  if (hasWCS())
    return wcs->imflip ? Frame::X : Frame::NORMAL;
  else
    return Frame::NORMAL;
}

int FitsImage::hasDATAMIN()
{
  return data ? data->hasDATAMIN() : 0;
}

int FitsImage::hasIRAFMIN()
{
  return data ? data->hasIRAFMIN() : 0;
}

void FitsImage::updateMinMax(Frame::MinMaxMode mm, int mmi)
{
  if (data)
    data->updateMinMax(mm,mmi);
}

void FitsImage::updateClip(Frame::ClipMode cm, Frame::MinMaxMode mm, float zc, 
			   int zs, int zl, int mmi, double ul, 
			   double uh, int fs)
{
  if (data)
    data->updateClip(cm,mm,zc,zs,zl,mmi,ul,uh,fs);
}

const char* FitsImage::getLow()
{
  return data ? data->getLow() : NULL;
}

const char* FitsImage::getHigh()
{
  return data ? data->getHigh() : NULL;
}

double FitsImage::getLowDouble()
{
  return data ? data->getLowDouble() : 0;
}

double FitsImage::getHighDouble()
{
  return data ? data->getHighDouble() : 0;
}

const char* FitsImage::getMin()
{
  return data ? data->getMin() : NULL;
}

const char* FitsImage::getMax()
{
  return data ? data->getMax() : NULL;
}

double FitsImage::getMinDouble()
{
  return data ? data->getMinDouble() : 0;
}

double FitsImage::getMaxDouble()
{
  return data ? data->getMaxDouble() : 0;
}
