#include "frame.h"
#include "fitsimage.h"

Vector Frame::map(const Vector& v, CoordSystem in, CoordSystem out, int* valid)
{
  if (valid) // assume we will succeed
    *valid = 1;

  if (in==out)
    return v;

  switch (in) {
  case NATIVEWCS:
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    return mapWCS(v,in,out,valid);

  case IMAGE:
    return mapImage(v,out,valid);
  case PHYSICAL:
    return mapPhysical(v,out,valid);
  case AMPLIFIER:
    return mapAmplifier(v,out,valid);
  case DETECTOR:
    return mapDetector(v,out,valid);
  case REF:
    return mapRef(v,out,valid);
  case USER:
    return mapUser(v,out,valid);
  case CANVAS:
    return mapCanvas(v,out,valid);
  case PANNER:
    return mapPanner(v,out,valid);

  default:
    cerr << "Frame Internal Error: unable to map invalid coordinate system." 
	 << endl;
  }

  // if we get here, we have an error condition

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapWCS(const Vector& v, CoordSystem in, CoordSystem out, 
		     int* valid)
{
  if (currentFits) {
    if (in!=NATIVEWCS) {
      switch (out) {
      case IMAGE:
	return currentFits->wcs2pix(v, in);
      case REF:
	return currentFits->wcs2pix(v, in) * currentFits->getImageToRef();
      case USER:
	return currentFits->wcs2pix(v, in) * currentFits->getImageToUser();
      default:
	cerr << "Frame Internal Error: unable to map WCS." << endl;
      }
    }
    else {
      switch (out) {
      case IMAGE:
	return currentFits->wcs2pix(v);
      case REF:
	return currentFits->wcs2pix(v) * currentFits->getImageToRef();
      case USER:
	return currentFits->wcs2pix(v) * currentFits->getImageToUser();
      default:
	cerr << "Frame Internal Error: unable to map WCS." << endl;
      }
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapImage(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getImageToRef();
    case USER:
      return v * currentFits->getImageToUser();

    default:
      cerr << "Frame Internal Error: unable to map IMAGE." << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapPhysical(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getPhysicalToRef();
    case USER:
      return v * currentFits->getPhysicalToUser();

    default:
      cerr << "Frame Internal Error: unable to map Physical" << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapAmplifier(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getAmplifierToRef();
    case USER:
      return v * currentFits->getAmplifierToUser();

    default:
      cerr << "Frame Internal Error: unable to map AMPLIFIER." << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapDetector(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getDetectorToRef();
    case USER:
      return v * currentFits->getDetectorToUser();

    default:
      cerr << "Frame Internal Error: unable to map DETECTOR." << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapRef(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case USER:
    return v * refToUser;
  case WIDGET:
    return v * refToWidget;
  case CANVAS:
    return v * refToCanvas;
  case WINDOW:
    return v *refToWindow;

  case IMAGE:
    if (currentFits)
      return v * currentFits->getRefToImage();
    break;
  case PHYSICAL:
    if (currentFits)
      return v * currentFits->getRefToPhysical();
    break;
  case AMPLIFIER:
    if (currentFits)
      return v * currentFits->getRefToAmplifier();
    break;
  case DETECTOR:
    if (currentFits)
      return v * currentFits->getRefToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getRefToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getRefToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map REF." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}      

Vector Frame::mapUser(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case REF:
    return v * userToRef;
  case WIDGET:
    return v * userToWidget;
  case CANVAS:
    return v * userToCanvas;

  case IMAGE:
    if (currentFits)
      return v * currentFits->getUserToImage();
    break;
  case PHYSICAL:
    if (currentFits)
      return v * currentFits->getUserToPhysical();
    break;
  case AMPLIFIER:
    if (currentFits)
      return v * currentFits->getUserToAmplifier();
    break;
  case DETECTOR:
    if (currentFits)
      return v * currentFits->getUserToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getUserToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getUserToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map USER." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapCanvas(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case REF:
    return v * canvasToRef;
  case USER:
    return v * canvasToUser;
  case WIDGET:
    return v * canvasToWidget;
  case WINDOW:
    return v * canvasToWindow;

  case IMAGE:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  return v * ptr->getCanvasToImage();

	ptr = ptr->next();
      }
    }
    break;
  case PHYSICAL:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasLTMV())
	    return v * ptr->getCanvasToPhysical();

	ptr = ptr->next();
      }
    }
    break;
  case AMPLIFIER:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasATMV() && ptr->hasLTMV())
	    return v * ptr->getCanvasToAmplifier();

	ptr = ptr->next();
      }
    }
    break;
  case DETECTOR:
    if (currentFits && currentFits->hasDTMV() && currentFits->hasLTMV())
      return v * currentFits->getCanvasToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getCanvasToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map CANVAS." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapPanner(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case REF:
    return v * pannerToRef;
  case USER:
    return v * pannerToUser;

  case IMAGE:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  return v * ptr->getPannerToImage();

	ptr = ptr->next();
      }
    }
    break;
  case PHYSICAL:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasLTMV())
	    return v * ptr->getPannerToPhysical();
	
	ptr = ptr->next();
      }
    }
    break;
  case AMPLIFIER:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasATMV() && ptr->hasLTMV())
	    return v * ptr->getPannerToAmplifier();

	ptr = ptr->next();
      }
    }
    break;
  case DETECTOR:
    if (currentFits && currentFits->hasDTMV() && currentFits->hasLTMV())
      return v * currentFits->getPannerToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]< ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getPannerToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map PANNER." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}

void Frame::mapToWCSStr(const Vector& v, CoordSystem in, CoordSystem out, 
			CoordFormat format, char* buf, int length)
{
  switch (in) {
  case REF:
    mapMatrixToWCSStr(v, out, format, buf, length,
		      &FitsImage::getRefToData, &FitsImage::getRefToImage);
    return;
  case USER:
    mapMatrixToWCSStr(v, out, format, buf, length,
		      &FitsImage::getUserToData, &FitsImage::getUserToImage);
    return;
  case CANVAS:
    mapMatrixToWCSStr(v, out, format, buf, length, 
		      &FitsImage::getCanvasToData, 
		      &FitsImage::getCanvasToImage);
    return;
  case PANNER:
    mapMatrixToWCSStr(v, out, format, buf, length,
		      &FitsImage::getPannerToData, 
		      &FitsImage::getPannerToImage);
    return;

  default:
    cerr << "Frame Internal Error: unable to mapToWCSStr invalid " 
	 << "coordinate system." << endl;
  }

  // if we get here, we have an error condition

  strcpy(buf,"");
}

void Frame::mapMatrixToWCSStr(const Vector& v, CoordSystem wcs, 
			      CoordFormat format, char* buf, int length,
			      Matrix& (FitsImage::*getMatrixData)(),
			      Matrix& (FitsImage::*getMatrixImage)())
{
  if (fits) {
    if (fitsCount==1) { // only one wcs to check
      fits->pix2wcs(v * (fits->*getMatrixImage)(), wcs, format, buf, length);
      return;
    }
    else { // check all available wcs
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * (ptr->*getMatrixData)();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height()) {
	  ptr->pix2wcs(v * (ptr->*getMatrixImage)(), wcs, format, buf, length);
	  return;
	}
	ptr = ptr->next();
      }

      // ok, its not on any images, use the first wcs
      fits->pix2wcs(v * (fits->*getMatrixImage)(), wcs, format, buf, length);
      return;
    }
  }

  // if we get here, we have an error condition

  strcpy(buf,"");
}

// mapDist

Vector Frame::mapDistToRef(const Vector& v, CoordSystem sys, 
			   CoordFormat format)
{
  // if format is DEGREES,ARCMIN,ARCSEC, do wcs. if PIXEL, look at coord system

  switch (format) {
  case PIXELS:
    switch (sys) {
    case NATIVEWCS:
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case IMAGE:
    case CANVAS:
      return v;

    case PHYSICAL:
      if (currentFits && currentFits->hasLTMV())
	return v * (currentFits->getPhysicalToImage()).abs();
      break;
    case AMPLIFIER:
      if (currentFits && currentFits->hasATMV())
	return v * (currentFits->getAmplifierToPhysical()*
		    currentFits->getPhysicalToImage()).abs();
      break;
    case DETECTOR:
      if (currentFits && currentFits->hasDTMV())
	return v * (currentFits->getDetectorToPhysical()*
		    currentFits->getPhysicalToImage()).abs();
      break;
    default:
      cerr << "Frame Internal Error: mapDistToRef." << endl;
    }
    break;

  case DEGREES:
    if (currentFits && currentFits->hasWCS())
      return currentFits->wcs2pixDist(v);
    break;
  case ARCMIN:
    if (currentFits && currentFits->hasWCS())
      return currentFits->wcs2pixDist(v/60);
    break;
  case ARCSEC:
    if (currentFits && currentFits->hasWCS())
      return currentFits->wcs2pixDist(v/(60*60));
    break;
  }

  // error

  return v;
}

Vector Frame::mapDistFromRef(const Vector& v, CoordSystem sys, 
			     CoordFormat format)
{
  // if format is DEGREES,ARCMIN,ARCSEC, do wcs. if PIXEL, look at coord system

  switch (format) {
  case PIXELS:
    switch (sys) {
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case IMAGE:
    case CANVAS:
      return v;

    case PHYSICAL:
      if (currentFits && currentFits->hasLTMV())
	return v * (currentFits->getImageToPhysical()).abs();
      break;
    case AMPLIFIER:
      if (currentFits && currentFits->hasATMV())
	return v * (currentFits->getImageToPhysical()*
		    currentFits->getPhysicalToAmplifier()).abs();
      break;
    case DETECTOR:
      if (currentFits && currentFits->hasDTMV())
	return v * (currentFits->getImageToPhysical()*
		    currentFits->getPhysicalToDetector()).abs();
      break;

    default:
      cerr << "Frame Internal Error: mapDistFromRef" << endl;
    }
    break;

  case DEGREES:
    if (currentFits && currentFits->hasWCS())
      return currentFits->pix2wcsDist(v);
    break;
  case ARCMIN:
    if (currentFits && currentFits->hasWCS())
      return currentFits->pix2wcsDist(v*60);
    break;
  case ARCSEC:
    if (currentFits && currentFits->hasWCS())
      return currentFits->pix2wcsDist(v*60*60);
    break;
  }

  // error

  return v;
}

double Frame::mapDistFromRef(const Vector& v1, const Vector& v2, 
			     CoordSystem sys, CoordFormat format)
{
  // if format is DEGREES,ARCMIN,ARCSEC, do wcs. if PIXEL, look at coord system

  switch (format) {
  case PIXELS:
    switch (sys) {
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case IMAGE:
    case CANVAS:
      return (v2-v1).length();

    case PHYSICAL:
      if (currentFits && currentFits->hasLTMV()) {
	Vector a = v1 * currentFits->getImageToPhysical();
	Vector b = v2 * currentFits->getImageToPhysical();
	return (b-a).length();
      }
      break;
    case AMPLIFIER:
      if (currentFits && currentFits->hasATMV()) {
	Vector a = v1 * currentFits->getImageToPhysical()*
	  currentFits->getPhysicalToAmplifier();
	Vector b = v2 * currentFits->getImageToPhysical()*
	  currentFits->getPhysicalToAmplifier();
	return (b-a).length();
      }
      break;
    case DETECTOR:
      if (currentFits && currentFits->hasDTMV()) {
	Vector a = v1 * currentFits->getImageToPhysical()*
	  currentFits->getPhysicalToDetector();
	Vector b = v2 * currentFits->getImageToPhysical()*
	  currentFits->getPhysicalToDetector();
	return (b-a).length();
      }
      break;

    default:
      cerr << "Frame Internal Error: mapDistFromRef" << endl;
    }
    break;

  case DEGREES:
    if (currentFits && currentFits->hasWCS())
      return wcsdist(v1,v2);
    break;
  case ARCMIN:
    if (currentFits && currentFits->hasWCS())
      return wcsdist(v1,v2)*60;
    break;
  case ARCSEC:
    if (currentFits && currentFits->hasWCS())
      return wcsdist(v1,v2)*60*60;
    break;
  }

  // error

  return 0;
}

double Frame::wcsdist(const Vector& v1, const Vector& v2)
{
  // v1 and v2 are in REF coords

  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v1 * ptr->getRefToData();
      if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	  img[1]<ptr->height()) {
	if (ptr->hasWCS()) {
	  Vector a = ptr->pix2wcs(v1 * ptr->getRefToImage());
	  Vector b = ptr->pix2wcs(v2 * ptr->getRefToImage());
	  return ptr->wcsdist(a,b);
	}
	else
	  return 0;
      }
      ptr = ptr->next();
    }
  }

  // if we get here, we have an error condition

  return 0;
}

void Frame::listCoordSystem(ostream& str, CoordSystem sys)
{
  switch (sys) {
  case IMAGE:
    str << "image";
    return;
  case PHYSICAL:
    str << "physical";
    return;
  case DETECTOR:
    str << "detector";
    return;
  case AMPLIFIER:
    str << "amplifier";
    return;
  case EQUATORIAL_FK4:
    str << "fk4";
    return;
  case EQUATORIAL_FK5:
    str << "fk5";
    return;
  case ICRS:
    str << "icrs";
    return;
  case GALACTIC:
    str << "galactic";
    return;
  case ECLIPTIC:
    str << "ecliptic";
    return;
  case LINEAR:
    str << "linear";
    return;
  }
}

void Frame::listCoordFormat(ostream& str, CoordFormat format)
{
  switch (format) {
  case PIXELS:
    str << "pixels";
    return;
  case DEGREES:
    str << "degrees";
    return;
  case SEXAGESIMAL:
    str << "sexagesimal";
    return;
  case ARCMIN:
    str << "arcmin";
    return;
  case ARCSEC:
    str << "arcsec";
    return;
  }
}

void Frame::listProsCoordSystem(ostream& str, CoordSystem sys)
{
  switch (sys) {
  case IMAGE:
    str << "logical";
    return;
  case PHYSICAL:
    str << "physical";
    return;
  case DETECTOR:
    str << "detector";
    return;
  case AMPLIFIER:
    str << "amplifier";
    return;
  case EQUATORIAL_FK4:
    str << "b1950";
    return;
  case EQUATORIAL_FK5:
    str << "j2000";
    return;
  case ICRS:
    str << "icrs";
    return;
  case GALACTIC:
    str << "galactic";
    return;
  case ECLIPTIC:
    str << "ecliptic";
    return;
  case LINEAR:
    str << "linear";
    return;
  }
}




