%{
#define YYPARSE_PARAM fr
#define FR ((Frame*)fr)
#define YYDEBUG 1

#define DISCARD_(x) {yyclearin; mkDiscard(x);}

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "util.h"
#include "vector.h"
#include "list.h"

extern int mklex(void);
extern void mkerror(const char*);
extern void mkDiscard(int);

static Frame::CoordSystem globalSystem;
static Frame::CoordSystem localSystem;

static Frame::CoordFormat localFormat;

static int sexFlag;
static int sexSign1;
static int sexSign2;
static char sexStr1[32];
static char sexStr2[32];

static unsigned short globalProps;
static unsigned short localProps;

static char globalColor[16];
static char localColor[16];

static char globalFont[32];
static char localFont[32];

static char globalText[80];
static char localText[80];

static Vector localVector;
static Vector localVector2;

static int globalLine1;
static int localLine1;
static int globalLine2;
static int localLine2;

static int globalPoint;
static int localPoint;

static Frame::CoordFormat globalRuler;
static Frame::CoordFormat localRuler;

static List<Vertex>* polylist;

static int tngText;
static int tngHeader;

static void setProps(unsigned short* props, unsigned short prop, int value);
static void setWCSSystem();
static void createPoint();

enum {CIRCLE,BOX,DIAMOND,CROSS,X,ARROW,BOXCIRCLE};
	
%}

%union {
  float real;
  int integer;
  char str[256];
}

%type <real> numeric
%type <integer> yesno

%type <real> angle
%type <real> optangle
%type <real> value
%type <integer> property
%type <integer> coordFormat
%type <integer> coordSystem
%type <integer> wcsSystem
%type <integer> equatorial
%type <integer> pointProp
%type <integer> numberof

%token <integer> INT
%token <real> REAL
%token <str> STRING

%token <integer> HOUR
%token <integer> MINUTE
%token <real> SECOND

%token <real> DEGREE
%token <real> ARCMINUTE
%token <real> ARCSECOND

%token <str> SEXSTR

%token <real> RADIAN
%token <integer> PIXEL

%token AMPLIFIER_
%token ANNULUS_
%token ARCMIN_
%token ARCSEC_
%token ARROW_
%token B1950_
%token BOX_
%token BOXCIRCLE_
%token CIRCLE_
%token COLOR_
%token COORDSYS_
%token CROSS_
%token DATE_
%token DEBUG_
%token DEGREES_
%token DELETE_
%token DETECTOR_
%token DIAMOND_
%token ECLIPTIC_
%token EDIT_
%token ELLIPSE_
%token EQUATORIAL_
%token FALSE_
%token FIELD_
%token FILENAME_
%token FIXED_
%token FK4_
%token FK5_
%token FONT_
%token FORMAT_
%token GALACTIC_
%token GLOBAL_
%token HMS_
%token ICRS_
%token IMAGE_
%token INCLUDE_
%token J2000_
%token LINE_
%token LINEAR_
%token LOCAL_
%token LOGICAL_
%token MOVE_
%token N_
%token NO_
%token OFF_
%token ON_
%token PANDA_
%token PHYSICAL_
%token PIE_
%token PIXELS_
%token POINT_
%token POLYGON_
%token RECTANGLE_
%token ROTATE_
%token ROTBOX_
%token RULER_
%token SELECT_
%token TEXT_
%token TRUE_
%token VERSION_
%token X_
%token Y_
%token YES_

%%

start	: initGlobal initLocal commands
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "DS9 Regions File 2.0" << endl;}
	| GLOBAL_ global eolComment
	| oldglobal eolComment /* old ds9 */
	| coordSystem /* pros, old ds9 */
	  {globalSystem = localSystem = (Frame::CoordSystem)$1;}
	| LOCAL_ local eolComment
	| include shape initLocal
	| comment
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

yesno	: INT {$$=($1 ? 1 : 0);}

	| YES_ {$$=1;}
	| Y_ {$$=1;}
	| ON_ {$$=1;}
	| TRUE_ {$$=1;}

	| NO_ {$$=0;}
	| N_ {$$=0;}
	| OFF_ {$$=0;}
	| FALSE_ {$$=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

optangle: /* empty */ {$$ = 0;}
	| angle {$$ = $1;}
	;

angle: numeric {$$ = degToRad($1);} /* assume degree */
	| DEGREE {$$ = degToRad($1);} /* assume degree */
	| RADIAN {$$=$1;}
	;

value	: numeric {$$=$1; localFormat = Frame::PIXELS;}
	| PIXEL   {$$=$1; localFormat = Frame::PIXELS;}
	| DEGREE  {$$=$1; localFormat = Frame::DEGREES;}
	| ARCMINUTE {$$=$1; localFormat = Frame::ARCMIN;}
	| ARCSECOND {$$=$1; localFormat = Frame::ARCSEC;}
	;

numberof: 'n' '=' INT {$$ = $3;}
	;

coord	: SEXSTR {strcpy(sexStr1,$1);} sp SEXSTR {strcpy(sexStr2,$4);} 
	{
	  sexFlag = 1;
	  localVector = Vector(FR->parseDMS(sexStr1), FR->parseDMS(sexStr2));
	  setWCSSystem();
	}
	| HOUR {sexSign1 = sexSign;} MINUTE SECOND sp
	  INT {sexSign2 = sexSign;} ARCMINUTE ARCSECOND
        {
	  sexFlag = 0;
	  localVector = Vector(hmsToDegree(sexSign1,fabs($1),$3,$4),
	    dmsToDegree(sexSign2,fabs($6),$8,$9));
	  setWCSSystem();
	}
	| INT {sexSign1 = sexSign;} ARCMINUTE ARCSECOND sp
	  INT {sexSign2 = sexSign;} ARCMINUTE ARCSECOND
        {
	  sexFlag = 0;
	  localVector = Vector(dmsToDegree(sexSign1,fabs($1),$3,$4),
	    dmsToDegree(sexSign,fabs($6),$8,$9));
	  setWCSSystem();
	}
	| numeric sp numeric
	{
	  sexFlag = 0;
	  localVector = Vector($1,$3);
	}
	| PIXEL sp PIXEL 
	{
	  sexFlag = 0;
	  localVector = Vector($1,$3);
	}
	| DEGREE sp DEGREE 
	{
	  sexFlag = 0;
	  localVector = Vector($1,$3);
	  setWCSSystem();
	}
	;

coordSystem : IMAGE_ {$$ = Frame::IMAGE;}
	| LOGICAL_ {$$ = Frame::IMAGE;}
	| PHYSICAL_ {$$ = Frame::PHYSICAL;}
	| AMPLIFIER_ {$$ = Frame::AMPLIFIER;}
	| DETECTOR_ {$$ = Frame::DETECTOR;}
	| wcsSystem {$$ = $1;}
	;

wcsSystem : EQUATORIAL_ equatorial {$$ = $2;}
	| FK4_ {$$ = Frame::EQUATORIAL_FK4;}
	| FK5_ {$$ = Frame::EQUATORIAL_FK5;}
	| B1950_  {$$ = Frame::EQUATORIAL_FK4;}
	| J2000_  {$$ = Frame::EQUATORIAL_FK5;}	
	| GALACTIC_ {$$ = Frame::GALACTIC;}
	| ECLIPTIC_ {$$ = Frame::ECLIPTIC;}
	| LINEAR_ {$$ = Frame::LINEAR;}
	| ICRS_ {$$ = Frame::ICRS;}
	;

equatorial : B1950_  {$$ = Frame::EQUATORIAL_FK4;}
	| J2000_  {$$ = Frame::EQUATORIAL_FK5;}
	| FK4_  {$$ = Frame::EQUATORIAL_FK4;}
	| FK5_  {$$ = Frame::EQUATORIAL_FK5;}
	;

coordFormat : PIXELS_ {$$=Frame::PIXELS;}
	| DEGREES_ {$$=Frame::DEGREES;}
	| ARCMIN_ {$$=Frame::ARCMIN;}
	| ARCSEC_ {$$=Frame::ARCSEC;}
	;

property : SELECT_ {$$ = Marker::SELECT;}
	| EDIT_ {$$ = Marker::EDIT;}
	| MOVE_  {$$ = Marker::MOVE;}
	| ROTATE_ {$$ = Marker::ROTATE;}
	| DELETE_ {$$ = Marker::DELETE;}
	| INCLUDE_ {$$ = Marker::INCLUDE;}
	| FIXED_ {$$ = Marker::FIXED;}
	;

global	: global sp globalProperty
	| globalProperty
	;

globalProperty : property '=' yesno 
	{
	  setProps(&globalProps,$1,$3);
	  setProps(&localProps,$1,$3);
	}
	| COLOR_ '=' STRING 
	{
	  strncpy(globalColor,$3,16);
	  strncpy(localColor,$3,16);
	}
	| FONT_ '=' STRING 
	{
	  strncpy(globalFont,$3,32);
	  strncpy(localFont,$3,32);
	}
	| TEXT_ '=' STRING 
	{
	  strncpy(globalText,$3,80);
	  strncpy(localText,$3,80);
	}
	| COORDSYS_ '=' coordSystem 
	{
	  globalSystem = localSystem = (Frame::CoordSystem)$3;
	}
	| POINT_ '=' pointProp 
	{
	  globalPoint = localPoint = $3;
	}
	| LINE_ '=' INT INT 
	{
	  globalLine1 = localLine1 = $3;
	  globalLine2 = localLine2 = $4;
	}
	| RULER_ '=' coordFormat 
	{
	  globalRuler = localRuler = (Frame::CoordFormat)$3;
	}
	;

initGlobal:{
	  // global properties
	  globalSystem = Frame::IMAGE;
	  globalProps =
	    Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE;
	  strcpy(globalColor,"green");
	  strcpy(globalFont,"helvetica 10 normal");
	  strcpy(globalText,"");

	  // unique properties
	  globalLine1 = 0;
	  globalLine2 = 0;
	  globalRuler = Frame::PIXELS;
	  globalPoint = BOXCIRCLE;

	  // other stuff
	  tngText = 0;   // we have found a comment to be used as text
	  tngHeader = 0; // we have a tng header
	} 
	;

local	: local sp localProperty
	| localProperty
	;

localProperty : property '=' yesno {setProps(&localProps,$1,$3);}
	| COLOR_ '=' STRING {strncpy(localColor,$3,16);}
	| FONT_ '=' STRING {strncpy(localFont,$3,32);}
	| TEXT_ '=' STRING {strncpy(localText,$3,80);}
	| COORDSYS_ '=' coordSystem {localSystem = (Frame::CoordSystem)$3;}
	| POINT_ '=' pointProp {localPoint = $3;}
	| LINE_ '=' INT INT {localLine1=$3; localLine2=$4;}
	| RULER_ '=' coordFormat {localRuler = (Frame::CoordFormat)$3;}
	;

initLocal : {
	  // global properties
	  localSystem = globalSystem;
	  localProps = globalProps;
	  strcpy(localColor,globalColor);
	  strcpy(localFont,globalFont);
	  strcpy(localText,globalText);

	  // unique properties
	  localLine1 = globalLine1;
	  localLine2 = globalLine2;
	  localRuler = globalRuler;
	  localPoint = globalPoint;

	  // other stuff
	  // we need this for tng format text comments
	  if (tngText) {
	    strcpy(globalText,"");
	    tngText=0;
          }
	}
	;

pointProp : CIRCLE_ {$$ = CIRCLE;}
	| BOX_ {$$ = BOX;}
	| DIAMOND_ {$$ = DIAMOND;}
	| CROSS_ {$$ = CROSS;}
	| X_ {$$ = X;}
	| ARROW_ {$$ = ARROW;}
	| BOXCIRCLE_ {$$ = BOXCIRCLE;}
	;

include	: /* empty */ {setProps(&localProps, Marker::INCLUDE, 1);}
	| '+' {setProps(&localProps, Marker::INCLUDE, 1);}
	| '-' {setProps(&localProps, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep post shapeComment
	  {FR->createCircleCmd(localVector,localSystem,$5,localFormat,
	    localColor,localFont,localText,localProps);}

	| ELLIPSE_ bp coord sp value sp value sp optangle ep post
	    shapeComment
	  {FR->createEllipseCmd(localVector,localSystem,Vector($5,$7),
	    localFormat,$9,localColor,localFont,localText,
	    localProps);}

	| BOX_ bp coord sp value sp value sp optangle ep post
	    shapeComment
	  {FR->createBoxCmd(localVector,localSystem,Vector($5,$7),
	    localFormat,$9,localColor,localFont,localText,
	    localProps);}
	| RECTANGLE_ bp coord sp value sp value sp optangle ep post
	    shapeComment
	  {FR->createBoxCmd(localVector,localSystem,Vector($5,$7),
	    localFormat,$9,localColor,localFont,localText,
	    localProps);}
	| ROTBOX_ bp coord sp value sp value sp optangle ep post
	    shapeComment
	  {FR->createBoxCmd(localVector,localSystem,Vector($5,$7),
	    localFormat,$9,localColor,localFont,localText,
	    localProps);}

	| POLYGON_ {polylist=new List<Vertex>();} bp polyNodes ep polypost
	    shapeComment
	  {FR->createPolygonCmd(polylist, localSystem,
	    localColor,localFont,localText,localProps);} 

	| LINE_ bp coord {localVector2 = localVector;} sp coord ep post2
	    oldline shapeComment
	  {FR->createLineCmd(localVector2,localVector,localSystem,
	    localLine1,localLine2,localColor,localFont,localText,localProps);}

	| TEXT_ bp coord ep post shapeComment
	  {FR->createTextCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| TEXT_ bp coord sp STRING ep {strncpy(localText,$5,80);} post
	    shapeComment
	  {FR->createTextCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}

	| RULER_ bp coord {localVector2 = localVector;} sp coord ep 
	    post2 oldruler shapeComment
	  {FR->createRulerCmd(localVector2,localVector,localSystem,
	    localRuler,localColor,localFont,localText,localProps);}

	| POINT_ bp coord ep post shapeComment 
	{
	  switch (localPoint) {
	  case CIRCLE:
	    FR->createCirclePointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  case BOX:
	    FR->createBoxPointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  case DIAMOND:
	    FR->createDiamondPointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  case CROSS:
	    FR->createCrossPointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  case X:
	    FR->createExPointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  case ARROW:
	    FR->createArrowPointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  case BOXCIRCLE:
	    FR->createBoxCirclePointCmd(localVector,localSystem,
	      localColor,localFont,localText,localProps);
	    break;
	  }
	}
	| CIRCLE_ POINT_ bp coord ep post shapeComment
	  {FR->createCirclePointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| BOX_ POINT_ bp coord ep post shapeComment
	  {FR->createBoxPointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| DIAMOND_ POINT_ bp coord ep post shapeComment
	  {FR->createDiamondPointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| CROSS_ POINT_ bp coord ep post shapeComment
	  {FR->createCrossPointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| X_ POINT_ bp coord ep post shapeComment
	  {FR->createExPointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| ARROW_ POINT_ bp coord ep post shapeComment
	  {FR->createArrowPointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}
	| BOXCIRCLE_ POINT_ bp coord ep post shapeComment
	  {FR->createBoxCirclePointCmd(localVector,localSystem,
	    localColor,localFont,localText,localProps);}

	| ANNULUS_ bp coord sp value sp value ep post shapeComment
	| ANNULUS_ bp coord sp value sp value sp annulusRads ep 
	    post shapeComment
	| ANNULUS_ bp coord sp value sp value sp numberof ep post shapeComment
	| PIE_ bp coord sp angle sp angle ep post shapeComment
	| PIE_ bp coord sp angle sp angle sp pieAngs ep post shapeComment
	| PIE_ bp coord sp angle sp angle sp numberof ep post shapeComment
	| FIELD_ bp ep shapeComment
	| PANDA_ bp coord sp angle sp angle sp INT sp 
	    numeric sp numeric sp INT ep post shapeComment
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist->append(new Vertex(localVector));}
	;

annulusRads: annulusRads sp annulusRad
	| annulusRad
	;

annulusRad: value {}
	;

pieAngs	: pieAngs sp pieAng
	| pieAng
	;

pieAng	: angle {}
	;

post	 : /* empty */ fixcoord
	| /* empty */ fixcoord oldlocal
	| coordSystem {localSystem = (Frame::CoordSystem)$1;} fixcoord
	| coordSystem {localSystem = (Frame::CoordSystem)$1;} 
	    fixcoord oldlocal
	;

fixcoord : /* empty */
	{
	  if (sexFlag) 
	    // convert degree to hours for ra
	    switch (localSystem) {
	    case Frame::NATIVEWCS:
	    case Frame::EQUATORIAL_FK4:
	    case Frame::EQUATORIAL_FK5:
	    case Frame::ICRS:
	      localVector[0] *= 360./24.;	    
	    }
	}
	;

post2	: /* empty */ fixcoord2
	| /* empty */ fixcoord2 oldlocal
	| coordSystem {localSystem = (Frame::CoordSystem)$1;} fixcoord2
	| coordSystem {localSystem = (Frame::CoordSystem)$1;}
	  fixcoord2 oldlocal
	;	

fixcoord2 : /* empty */
	{
	  if (sexFlag) 
	    // convert degree to hours for ra
	    switch (localSystem) {
	    case Frame::NATIVEWCS:
	    case Frame::EQUATORIAL_FK4:
	    case Frame::EQUATORIAL_FK5:
	    case Frame::ICRS:
	      localVector[0] *= 360./24.;	    
	      localVector2[0] *= 360./24.;	    
	    }
	}
	;

polypost: /* empty */ polyfixcoord
	| /* empty */ polyfixcoord oldlocal
	| coordSystem {localSystem = (Frame::CoordSystem)$1;} polyfixcoord
	| coordSystem {localSystem = (Frame::CoordSystem)$1;}
	  polyfixcoord oldlocal
	;

polyfixcoord : /* empty */
	{
	  if (sexFlag) 
	    // convert degree to hours for ra
	    switch (localSystem) {
	    case Frame::NATIVEWCS:
	    case Frame::EQUATORIAL_FK4:
	    case Frame::EQUATORIAL_FK5:
	    case Frame::ICRS:
	      polylist->head();
	      do
	        polylist->current()->vector[0] *= 360./24.;
	      while (polylist->next());
	    }
	}
	;

comment	: '#' FORMAT_ ':' {tngHeader=1;} tngFormat
	| '#' FILENAME_ ':' {tngHeader=1; DISCARD_(0)} STRING
	| '#' DATE_ ':' {tngHeader=1; DISCARD_(0)} STRING
	| '#' {DISCARD_(1)} STRING 
	{
	  /* we need this for tng format text comments */
	  if (tngHeader) {
	    strncpy(globalText,$3,80);
	    strncpy(localText,$3,80);
	    tngText=1;
	  }
	}
	| '#' ARROW_ {DISCARD_(0)} STRING
	;

eolComment : /* empty */
	| '#' {DISCARD_(0)} STRING
	;

shapeComment : /* empty */
	| '#' {DISCARD_(1)} STRING
	{ if (tngHeader) {
	    // try to feature out if this is a color, not a comment
  	    if (!strcmp($3,"# white"))
	      strcpy(localColor,"white");
	    else if (!strcmp($3,"# black"))
	      strcpy(localColor,"black");
	    else if (!strcmp($3,"# red"))
	      strcpy(localColor,"red");
	    else if (!strcmp($3,"# green"))
	      strcpy(localColor,"green");
	    else if (!strcmp($3,"# blue"))
	      strcpy(localColor,"blue");
	    else if (!strcmp($3,"# cyan"))
	      strcpy(localColor,"cyan");
	    else if (!strcmp($3,"# magneta"))
	      strcpy(localColor,"magneta");
	    else if (!strcmp($3,"# yellow"))
	      strcpy(localColor,"yellow");

	    else if (!strcmp($3,"# source"))
	      strcpy(localColor,"green");
	    else if (!strcmp($3,"# background"))
	      strcpy(localColor,"red");

	    else if (!strcmp($3,"# purple"))
	      strcpy(localColor,"magneta");
	    else if (!strcmp($3,"# darkslategray"))
	      strcpy(localColor,"white");
	  }
        }
	;

oldglobal: oldglobal sp oldglobalProperty
	| oldglobalProperty
	;

oldglobalProperty : property '=' yesno 
	{
	  setProps(&globalProps,$1,$3);
	  setProps(&localProps,$1,$3);
	}
	| COLOR_ '=' STRING 
	{
	  strncpy(globalColor,$3,16);
	  strncpy(localColor,$3,16);
	}
	| FONT_ '=' STRING 
	{
	  strncpy(globalFont,$3,32);
	  strncpy(localFont,$3,32);
	}
	;

oldlocal: oldlocal sp oldlocalProperty
	| oldlocalProperty
	;

oldlocalProperty : property '=' yesno {setProps(&localProps,$1,$3);}
	| COLOR_ '=' STRING {strncpy(localColor,$3,16);}
	| FONT_ '=' STRING {strncpy(localFont,$3,32);}
	| TEXT_ '=' STRING {strncpy(localText,$3,80);}
	;

oldline: /* empty */
	| ARROW_ INT INT {localLine1=$2; localLine2=$3;}
	;

oldruler: /* empty */
	|  coordFormat {localRuler = (Frame::CoordFormat)$1;}
	;

tngFormat : HMS_ tngWCS
	| DEGREES_ tngWCS
	| PIXELS_ 
	  {globalSystem = localSystem = Frame::IMAGE;}
	| PIXELS_ '(' PHYSICAL_ ')' 
	  {globalSystem = localSystem = Frame::PHYSICAL;}
	;

tngWCS	: /* empty */
	  {globalSystem = localSystem = Frame::EQUATORIAL_FK4;}
	| '(' wcsSystem ')' 
	  {globalSystem = localSystem = (Frame::CoordSystem)$2;}
	;

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}

static void setWCSSystem()
{
  // if we have not defined a specific WCS, set to Native
  switch (localSystem) {
    case (Frame::IMAGE):
    case (Frame::PHYSICAL):
    case (Frame::DETECTOR):
    case (Frame::AMPLIFIER):
      localSystem = (Frame::NATIVEWCS);
      break;
  }
}
